# coding=utf-8
from __future__ import print_function

import copy
import itertools


class Context(object):
    def __init__(self, **kwargs):
        self._dict = kwargs
        self.__dict__.update(
            {
                k: Context(**v) if isinstance(v, dict) else v
                for k, v in kwargs.items()
            }
        )

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def __len__(self):
        return len(self._dict)

    def __str__(self):
        return "<{}: {}>".format(self.__class__.__name__, self._dict)

    def __repr__(self):
        return str(self)

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()

    def to_dict(self):
        return copy.copy(self._dict)


class Global(Context):
    def set(self, key, value):
        setattr(self, key, value)
        self._dict[key] = value

    def update(self, globals={}, **kwargs):
        for key, value in itertools.chain(globals.items(), kwargs.items()):
            self.set(key, value)


g = Global()



# coding=utf-8
from __future__ import print_function

import cv2
import imageio
import numpy as np
import pandas as pd
import functools
import collections

from suanpan import asyncio, convert, path


def loadFromCsv(filepath, **kwargs):
    kwargs.setdefault("index_col", 0)
    return pd.read_csv(filepath, **kwargs)


def saveAsCsv(filepath, data, **kwargs):
    path.safeMkdirsForFile(filepath)
    data.to_csv(filepath, data, **kwargs)
    return filepath


def loadFromNpy(filepath, **kwargs):
    kwargs.setdefault("encoding", "latin1")
    return np.load(filepath, **kwargs)


def saveAsNpy(filepath, data, **kwargs):
    path.safeMkdirsForFile(filepath)
    kwargs.setdefault("fix_imports", True)
    np.save(filepath, data, **kwargs)
    return filepath


def saveImage(filepath, image):
    path.safeMkdirsForFile(filepath)
    cv2.imwrite(filepath, image)  # pylint: disable-msg=e1101
    return filepath


def saveAsFlatImage(filepath, data):
    image = convert.flatAsImage(data)
    return saveImage(filepath, image)


def saveAsAnimatedGif(filepath, data):
    image3D = convert.to3D(data)
    path.safeMkdirsForFile(filepath)
    imageio.mimsave(filepath, image3D)
    return filepath


def saveAsImage(filepath, data, flag=None):
    mapping = {None: saveImage, "flat": saveAsFlatImage, "animated": saveAsAnimatedGif}
    func = mapping.get(flag)
    if not func:
        raise Exception("Unknow flag: {}".format(flag))
    return func(filepath, data)


def saveAsImages(filepathPrefix, images, workers=None):
    counts = len(images)
    n = len(str(counts))
    workers = workers or min(counts, asyncio.WORKERS)
    with asyncio.multiThread(workers) as pool:
        for index, image in enumerate(images):
            pool.apply_async(
                saveAsImage,
                args=("{}_{}.png".format(filepathPrefix, str(index).zfill(n)), image),
            )
    return filepathPrefix


def saveAllAsImages(filepathPrefix, data, workers=None, pool=None):
    def _save(filepathPrefix, image3D, pool):
        layers = len(image3D)
        n = len(str(layers))
        for index, image in enumerate(image3D):
            pool.apply_async(
                saveImage,
                args=("{}_{}.png".format(filepathPrefix, str(index).zfill(n)), image),
            )
        pool.apply_async(
            saveAsFlatImage, args=("{}.png".format(filepathPrefix), image3D)
        )
        pool.apply_async(
            saveAsAnimatedGif, args=("{}.gif".format(filepathPrefix), image3D)
        )

    image3D = convert.to3D(data)
    if pool:
        _save(filepathPrefix, image3D, pool)
    else:
        layers = len(image3D)
        workers = workers or min(layers + 2, asyncio.WORKERS)
        with asyncio.multiThread(workers) as pool:
            _save(filepathPrefix, image3D, pool)
    return filepathPrefix


def merge(*dicts):
    def _merge(result, item):
        result.update(item)
        return result

    return functools.reduce(_merge, dicts, {})


def count(iterable):
    if hasattr(iterable, '__len__'):
        return len(iterable)

    d = collections.deque(enumerate(iterable, 1), maxlen=1)
    return d[0][0] if d else 0
